API_CONTEXT = 'api/v1'
WEB_API_GROUP_PREFIX = 'webapi'
AUTHENTICATION_TYPE_BASIC = 'basic'
AUTHENTICATION_TYPE_OAUTH = 'oauth'
AUTHENTICATION_LONG_LIVING_TOKEN = 'apitoken'
AUTHENTICATION_TYPE_OAUTH_OFFLINE_TOKEN = 'oauth_token'
